@@ -1,7 +1,7 @@
 /*
     posix1e - a python module exposing the posix acl functions
 
-    Copyright (C) 2002-2009, 2012, 2014 Iustin Pop <iustin@k1024.org>
+    Copyright (C) 2002-2009, 2012, 2014, 2015 Iustin Pop <iustin@k1024.org>
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -45,7 +45,6 @@
 #define PyInt_AsUnsignedLongMask PyLong_AsUnsignedLongMask
 #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
 #define PyInt_AS_LONG PyLong_AS_LONG
-#define MyString_ConcatAndDel PyUnicode_AppendAndDel
 #define MyString_FromFormat PyUnicode_FromFormat
 #define MyString_FromString PyUnicode_FromString
 #define MyString_FromStringAndSize PyUnicode_FromStringAndSize
@@ -55,19 +54,34 @@
 #define PyBytes_FromStringAndSize PyString_FromStringAndSize
 #define PyBytes_FromString PyString_FromString
 #define PyBytes_FromFormat PyString_FromFormat
-#define PyBytes_ConcatAndDel PyString_ConcatAndDel
-#define MyString_ConcatAndDel PyBytes_ConcatAndDel
 #define MyString_FromFormat PyBytes_FromFormat
 #define MyString_FromString PyBytes_FromString
 #define MyString_FromStringAndSize PyBytes_FromStringAndSize
-
-/* Python 2.6 already defines Py_TYPE */
-#ifndef Py_TYPE
-#define Py_TYPE(o)    (((PyObject*)(o))->ob_type)
-#endif
 #endif
 
-static PyTypeObject ACL_Type;
+/* Used for cpychecker: */
+/* The checker automatically defines this preprocessor name when creating
+   the custom attribute: */
+#if defined(WITH_CPYCHECKER_TYPE_OBJECT_FOR_TYPEDEF_ATTRIBUTE)
+#define CPYCHECKER_TYPE_OBJECT_FOR_TYPEDEF(typename) \
+  __attribute__((cpychecker_type_object_for_typedef(typename)))
+#else
+/* This handles the case where we're compiling with a "vanilla"
+   compiler that doesn't supply this attribute: */
+#define CPYCHECKER_TYPE_OBJECT_FOR_TYPEDEF(typename)
+#endif
+
+/* The checker automatically defines this preprocessor name when creating
+   the custom attribute: */
+#if defined(WITH_CPYCHECKER_NEGATIVE_RESULT_SETS_EXCEPTION_ATTRIBUTE)
+   #define CPYCHECKER_NEGATIVE_RESULT_SETS_EXCEPTION \
+__attribute__((cpychecker_negative_result_sets_exception))
+   #else
+   #define CPYCHECKER_NEGATIVE_RESULT_SETS_EXCEPTION
+#endif
+
+static PyTypeObject ACL_Type
+  CPYCHECKER_TYPE_OBJECT_FOR_TYPEDEF("ACL_Object");
 static PyObject* ACL_applyto(PyObject* obj, PyObject* args);
 static PyObject* ACL_valid(PyObject* obj, PyObject* args);
 
@@ -77,8 +91,10 @@ static PyObject* ACL_set_state(PyObject *obj, PyObject* args);
 #endif
 
 #ifdef HAVE_LEVEL2
-static PyTypeObject Entry_Type;
-static PyTypeObject Permset_Type;
+static PyTypeObject Entry_Type
+  CPYCHECKER_TYPE_OBJECT_FOR_TYPEDEF("Entry_Object");
+static PyTypeObject Permset_Type
+  CPYCHECKER_TYPE_OBJECT_FOR_TYPEDEF("Permset_Object");
 static PyObject* Permset_new(PyTypeObject* type, PyObject* args,
                              PyObject *keywds);
 #endif
@@ -134,8 +150,8 @@ static int ACL_init(PyObject* obj, PyObject* args, PyObject *keywds) {
 #ifdef HAVE_LINUX
     static char *kwlist[] = { "file", "fd", "text", "acl", "filedef",
                               "mode", NULL };
-    char *format = "|etisO!sH";
-    mode_t mode = 0;
+    char *format = "|etisO!si";
+    int mode = -1;
 #else
     static char *kwlist[] = { "file", "fd", "text", "acl", "filedef", NULL };
     char *format = "|etisO!s";
@@ -177,7 +193,7 @@ static int ACL_init(PyObject* obj, PyObject* args, PyObject *keywds) {
     else if(filedef != NULL)
         self->acl = acl_get_file(filedef, ACL_TYPE_DEFAULT);
 #ifdef HAVE_LINUX
-    else if(PyMapping_HasKeyString(keywds, kwlist[5]))
+    else if(mode != -1)
         self->acl = acl_from_mode(mode);
 #endif
     else
@@ -343,7 +359,8 @@ static PyObject* ACL_richcompare(PyObject* o1, PyObject* o2, int op) {
         ret = n == 1 ? Py_True : Py_False;
         break;
     default:
-        ret = Py_NotImplemented;
+        PyErr_SetString(PyExc_TypeError, "ACLs are not orderable");
+        return NULL;
     }
     Py_INCREF(ret);
     return ret;
@@ -371,12 +388,14 @@ static PyObject* ACL_equiv_mode(PyObject* obj, PyObject* args) {
 }
 #endif
 
+#ifndef IS_PY3K
 /* Implementation of the compare for ACLs */
 static int ACL_nocmp(PyObject* o1, PyObject* o2) {
 
     PyErr_SetString(PyExc_TypeError, "cannot compare ACLs using cmp()");
     return -1;
 }
+#endif
 
 /* Custom methods */
 static char __applyto_doc__[] =
@@ -680,27 +699,31 @@ typedef struct {
     };
 } tag_qual;
 
+/* Pre-declaring the function is more friendly to cpychecker, sigh. */
+static int get_tag_qualifier(acl_entry_t entry, tag_qual *tq)
+  CPYCHECKER_NEGATIVE_RESULT_SETS_EXCEPTION;
+
 /* Helper function to get the tag and qualifier of an Entry at the
    same time. This is "needed" because the acl_get_qualifier function
    returns a pointer to different types, based on the tag value, and
    thus it's not straightforward to get the right type.
 
-   It sets a Python exception if an error occurs, and return 0 in this
-   case. If successful, the tag is set to the tag type, and the
+   It sets a Python exception if an error occurs, and returns -1 in
+   this case. If successful, the tag is set to the tag type, the
    qualifier (if any) to either the uid or the gid entry in the
-   tag_qual structure.
+   tag_qual structure, and the return value is 0.
 */
-int get_tag_qualifier(acl_entry_t entry, tag_qual *tq) {
+static int get_tag_qualifier(acl_entry_t entry, tag_qual *tq) {
     void *p;
 
     if(acl_get_tag_type(entry, &tq->tag) == -1) {
         PyErr_SetFromErrno(PyExc_IOError);
-        return 0;
+        return -1;
     }
     if (tq->tag == ACL_USER || tq->tag == ACL_GROUP) {
         if((p = acl_get_qualifier(entry)) == NULL) {
             PyErr_SetFromErrno(PyExc_IOError);
-            return 0;
+            return -1;
         }
         if (tq->tag == ACL_USER) {
             tq->uid = *(uid_t*)p;
@@ -709,7 +732,7 @@ int get_tag_qualifier(acl_entry_t entry, tag_qual *tq) {
         }
         acl_free(p);
     }
-    return 1;
+    return 0;
 }
 
 /* Creation of a new Entry instance */
@@ -769,7 +792,7 @@ static PyObject* Entry_str(PyObject *obj) {
     Entry_Object *self = (Entry_Object*) obj;
     tag_qual tq;
 
-    if(!get_tag_qualifier(self->entry, &tq)) {
+    if(get_tag_qualifier(self->entry, &tq) < 0) {
         return NULL;
     }
 
@@ -810,8 +833,15 @@ static PyObject* Entry_str(PyObject *obj) {
         Py_DECREF(format);
         return NULL;
     }
-    MyString_ConcatAndDel(&format, kind);
+#ifdef IS_PY3K
+    PyObject *ret = PyUnicode_Concat(format, kind);
+    Py_DECREF(format);
+    Py_DECREF(kind);
+    return ret;
+#else
+    PyString_ConcatAndDel(&format, kind);
     return format;
+#endif
 }
 
 /* Sets the tag type of the entry */
@@ -930,7 +960,7 @@ static PyObject* Entry_get_qualifier(PyObject *obj, void* arg) {
         PyErr_SetString(PyExc_AttributeError, "entry attribute");
         return NULL;
     }
-    if(!get_tag_qualifier(self->entry, &tq)) {
+    if(get_tag_qualifier(self->entry, &tq) < 0) {
         return NULL;
     }
     if (tq.tag == ACL_USER) {
@@ -1116,8 +1146,8 @@ static int Permset_set_right(PyObject* obj, PyObject* value, void* arg) {
     int nerr;
 
     if(!PyInt_Check(value)) {
-        PyErr_SetString(PyExc_ValueError, "a maximum of one argument must"
-                        " be passed");
+        PyErr_SetString(PyExc_ValueError, "invalid argument, an integer"
+                        " is expected");
         return -1;
     }
     on = PyInt_AsLong(value);
@@ -1230,17 +1260,20 @@ static char __ACL_Type_doc__[] =
     ".. note:: only one keyword parameter should be provided\n"
     "\n"
     ":param string file: creates an ACL representing\n"
-    "    the access ACL of the specified file\n"
+    "    the access ACL of the specified file.\n"
     ":param string filedef: creates an ACL representing\n"
-    "    the default ACL of the given directory\n"
+    "    the default ACL of the given directory.\n"
     ":param int fd: creates an ACL representing\n"
-    "    the access ACL of the given file descriptor\n"
+    "    the access ACL of the given file descriptor.\n"
     ":param string text: creates an ACL from a \n"
-    "    textual description\n"
-    ":param ACL acl: creates a copy of an existing ACL instance\n"
+    "    textual description; note the ACL must be valid, which\n"
+    "    means including a mask for extended ACLs, similar to\n"
+    "    ``setfacl --no-mask``\n"
+    ":param ACL acl: creates a copy of an existing ACL instance.\n"
     ":param int mode: creates an ACL from a numeric mode\n"
-    "    (e.g. mode=0644) (this is valid only when the C library\n"
-    "    provides the acl_from_mode call)\n"
+    "    (e.g. ``mode=0644``); this is valid only when the C library\n"
+    "    provides the ``acl_from_mode call``, and\n"
+    "    note that no validation is done on the given value.\n"
     "\n"
     "If no parameters are passed, an empty ACL will be created; this\n"
     "makes sense only when your OS supports ACL modification\n"
@@ -1288,7 +1321,12 @@ static PyTypeObject ACL_Type = {
     0,                  /* tp_print */
     0,                  /* tp_getattr */
     0,                  /* tp_setattr */
+#ifdef IS_PY3K
+    0,                  /* formerly tp_compare, in 3.0 deprecated, in
+                           3.5 tp_as_async */
+#else
     ACL_nocmp,          /* tp_compare */
+#endif
     0,                  /* tp_repr */
     0,                  /* tp_as_number */
     0,                  /* tp_as_sequence */
